package starling.animation {
import starling.events.EventDispatcher;
/**
 *  A Tween animates numeric properties of objects. It uses different transition functions
 *  *  to give the animations various styles.
 *  *  
 *  *  <p>The primary use of this class is to do standard animations like movement, fading, 
 *  *  rotation, etc. But there are no limits on what to animate; as long as the property you want
 *  *  to animate is numeric (<code>int, uint, Number</code>), the tween can handle it. For a list 
 *  *  of available Transition types, look at the "Transitions" class.</p> 
 *  *  
 *  *  <p>Here is an example of a tween that moves an object to the right, rotates it, and 
 *  *  fades it out:</p>
 *  *  
 *  *  <listing>
 *  *  var tween:Tween = new Tween(object, 2.0, Transitions.EASE_IN_OUT);
 *  *  tween.animate("x", object.x + 50);
 *  *  tween.animate("rotation", deg2rad(45));
 *  *  tween.fadeTo(0);    // equivalent to 'animate("alpha", 0)'
 *  *  Starling.juggler.add(tween);</listing> 
 *  *  
 *  *  <p>Note that the object is added to a juggler at the end of this sample. That's because a 
 *  *  tween will only be executed if its "advanceTime" method is executed regularly - the 
 *  *  juggler will do that for you, and will remove the tween when it is finished.</p>
 *  *  
 *  *  @see Juggler
 *  *  @see Transitions
 *  
 * @externs
 */
public class Tween extends starling.events.EventDispatcher implements starling.animation.IAnimatable {
	/**
	 *  Creates a tween with a target, duration (in seconds) and a transition function.
	 *      * @param target the object that you want to animate
	 *      * @param time the duration of the Tween (in seconds)
	 *      * @param transition can be either a String (e.g. one of the constants defined in the
	 *      *        Transitions class) or a function. Look up the 'Transitions' class for a   
	 *      *        documentation about the required function signature. 
	 */
	public function Tween(target:*, time:Number, transition:* = undefined) {
		super();
	}
	/**
	 *  Resets the tween to its default values. Useful for pooling tweens. 
	 */
	public function reset(target:*, time:Number, transition:* = undefined):starling.animation.Tween { return null; }
	/**
	 *  Animates the property of the target to a certain value. You can call this method
	 *      * multiple times on one tween.
	 *      *
	 *      * <p>Some property types are handled in a special way:</p>
	 *      * <ul>
	 *      *   <li>If the property contains the string <code>color</code> or <code>Color</code>,
	 *      *       it will be treated as an unsigned integer with a color value
	 *      *       (e.g. <code>0xff0000</code> for red). Each color channel will be animated
	 *      *       individually.</li>
	 *      *   <li>The same happens if you append the string <code>#rgb</code> to the name.</li>
	 *      *   <li>If you append <code>#rad</code>, the property is treated as an angle in radians,
	 *      *       making sure it always uses the shortest possible arc for the rotation.</li>
	 *      *   <li>The string <code>#deg</code> does the same for angles in degrees.</li>
	 *      * </ul>
	 *      
	 */
	public function animate(property:String, endValue:Number):void {}
	/**
	 *  Animates the 'scaleX' and 'scaleY' properties of an object simultaneously. 
	 */
	public function scaleTo(factor:Number):void {}
	/**
	 *  Animates the 'x' and 'y' properties of an object simultaneously. 
	 */
	public function moveTo(x:Number, y:Number):void {}
	/**
	 *  Animates the 'alpha' property of an object to a certain target value. 
	 */
	public function fadeTo(alpha:Number):void {}
	/**
	 *  Animates the 'rotation' property of an object to a certain target value, using the
	 *      * smallest possible arc. 'type' may be either 'rad' or 'deg', depending on the unit of
	 *      * measurement. 
	 */
	public function rotateTo(angle:Number, type:String = undefined):void {}
	/**
	 *  @inheritDoc 
	 */
	public function advanceTime(time:Number):void {}
	/**
	 *  The end value a certain property is animated to. Throws an ArgumentError if the 
	 *      * property is not being animated. 
	 */
	public function getEndValue(property:String):Number { return 0; }
	/**
	 *  Indicates if a property with the given name is being animated by this tween. 
	 */
	public function animatesProperty(property:String):Boolean { return false; }
	/**
	 *  Indicates if the tween is finished. 
	 */
	public function get isComplete():Boolean { return false; }
	/**
	 *  The target object that is animated. 
	 */
	public function get target():* { return null; }
	/**
	 *  The transition method used for the animation. @see Transitions 
	 */
	public function get transition():String { return null; }
	public function set transition(value:String):void {}
	/**
	 *  The actual transition function used for the animation. 
	 */
	public function get transitionFunc():Function { return null; }
	public function set transitionFunc(value:Function):void {}
	/**
	 *  The total time the tween will take per repetition (in seconds). 
	 */
	public function get totalTime():Number { return 0; }
	/**
	 *  The time that has passed since the tween was created (in seconds). 
	 */
	public function get currentTime():Number { return 0; }
	/**
	 *  The current progress between 0 and 1, as calculated by the transition function. 
	 */
	public function get progress():Number { return 0; }
	/**
	 *  The delay before the tween is started (in seconds). @default 0 
	 */
	public function get delay():Number { return 0; }
	public function set delay(value:Number):void {}
	/**
	 *  The number of times the tween will be executed. 
	 *      * Set to '0' to tween indefinitely. @default 1 
	 */
	public function get repeatCount():int { return 0; }
	public function set repeatCount(value:int):void {}
	/**
	 *  The amount of time to wait between repeat cycles (in seconds). @default 0 
	 */
	public function get repeatDelay():Number { return 0; }
	public function set repeatDelay(value:Number):void {}
	/**
	 *  Indicates if the tween should be reversed when it is repeating. If enabled, 
	 *      * every second repetition will be reversed. @default false 
	 */
	public function get reverse():Boolean { return false; }
	public function set reverse(value:Boolean):void {}
	/**
	 *  Indicates if the numeric values should be cast to Integers. @default false 
	 */
	public function get roundToInt():Boolean { return false; }
	public function set roundToInt(value:Boolean):void {}
	/**
	 *  A function that will be called when the tween starts (after a possible delay). 
	 */
	public function get onStart():Function { return null; }
	public function set onStart(value:Function):void {}
	/**
	 *  A function that will be called each time the tween is advanced. 
	 */
	public function get onUpdate():Function { return null; }
	public function set onUpdate(value:Function):void {}
	/**
	 *  A function that will be called each time the tween finishes one repetition
	 *      * (except the last, which will trigger 'onComplete'). 
	 */
	public function get onRepeat():Function { return null; }
	public function set onRepeat(value:Function):void {}
	/**
	 *  A function that will be called when the tween is complete. 
	 */
	public function get onComplete():Function { return null; }
	public function set onComplete(value:Function):void {}
	/**
	 *  The arguments that will be passed to the 'onStart' function. 
	 */
	public function get onStartArgs():Array { return null; }
	public function set onStartArgs(value:Array):void {}
	/**
	 *  The arguments that will be passed to the 'onUpdate' function. 
	 */
	public function get onUpdateArgs():Array { return null; }
	public function set onUpdateArgs(value:Array):void {}
	/**
	 *  The arguments that will be passed to the 'onRepeat' function. 
	 */
	public function get onRepeatArgs():Array { return null; }
	public function set onRepeatArgs(value:Array):void {}
	/**
	 *  The arguments that will be passed to the 'onComplete' function. 
	 */
	public function get onCompleteArgs():Array { return null; }
	public function set onCompleteArgs(value:Array):void {}
	/**
	 *  Another tween that will be started (i.e. added to the same juggler) as soon as 
	 *      * this tween is completed. 
	 */
	public function get nextTween():starling.animation.Tween { return null; }
	public function set nextTween(value:starling.animation.Tween):void {}
}
}
