package starling.assets {
import openfl.Vector;
/**
 *  An AssetFactory is responsible for creating a concrete instance of an asset.
 *  *
 *  *  <p>The AssetManager contains a list of AssetFactories, registered via 'registerFactory'.
 *  *  When the asset queue is processed, each factory (sorted by priority) will be asked if it
 *  *  can handle a certain AssetReference (via the 'canHandle') method. If it can, the 'create'
 *  *  method will be called, which is responsible for creating at least one asset.</p>
 *  *
 *  *  <p>By extending 'AssetFactory' and registering your class at the AssetManager, you can
 *  *  customize how assets are being created and even add new types of assets.</p>
 *  
 * @externs
 */
public class AssetFactory {
	/**
	 *  Creates a new instance. 
	 */
	function AssetFactory() {}
	/**
	 *  Returns 'true' if this factory can handle the given reference. The default
	 *      *  implementation checks if extension and/or mime type of the reference match those
	 *      *  of the factory. 
	 */
	public function canHandle(reference:starling.assets.AssetReference):Boolean { return false; }
	/**
	 *  This method will only be called if 'canHandle' returned 'true' for the given reference.
	 *      *  It's responsible for creating at least one concrete asset and passing it to 'onComplete'.
	 *      *
	 *      *  @param reference   The asset to be created. If a local or remote URL is referenced,
	 *      *                     it will already have been loaded, and 'data' will contain a ByteArray.
	 *      *  @param helper      Contains useful utility methods to be used by the factory. Look
	 *      *                     at the class documentation for more information.
	 *      *  @param onComplete  To be called when loading is successful. 'type' parameter is optional.
	 *      *                     <pre>function(name:String, asset:Object, type:String):void;</pre>
	 *      *  @param onError     To be called when creation fails for some reason. Do not call
	 *      *                     'onComplete' when that happens. <pre>function(error:String):void</pre>
	 *      
	 */
	public function create(reference:starling.assets.AssetReference, helper:starling.assets.AssetFactoryHelper, onComplete:Function, onError:Function):void {}
	/**
	 *  Add one or more mime types that identify the supported data types. Used by
	 *      *  'canHandle' to figure out if the factory is suitable for an asset reference. 
	 */
	public function addMimeTypes(args:Array):void {}
	/**
	 *  Add one or more file extensions (without leading dot) that identify the supported data
	 *      *  types. Used by 'canHandle' to figure out if the factory is suitable for an asset
	 *      *  reference. 
	 */
	public function addExtensions(args:Array):void {}
	/**
	 *  Returns the mime types this factory supports. 
	 */
	public function getMimeTypes(out:openfl.Vector = undefined):openfl.Vector { return null; }
	/**
	 *  Returns the file extensions this factory supports. 
	 */
	public function getExtensions(out:openfl.Vector = undefined):openfl.Vector { return null; }
	public function get priority():int { return 0; }
	public function set priority(value:int):void {}
}
}
