package starling.display {
import starling.rendering.Painter;
import starling.events.Event;
import openfl.geom.Rectangle;
import openfl.geom.Point;
/**
 * A DisplayObjectContainer represents a collection of display objects.
 * It is the base class of all display objects that act as a container for other objects. By 
 * maintaining an ordered list of children, it defines the back-to-front positioning of the 
 * children within the display tree.
 * 
 * <p>A container does not a have size in itself. The width and height properties represent the 
 * extents of its children. Changing those properties will scale all children accordingly.</p>
 * 
 * <p>As this is an abstract class, you can't instantiate it directly, but have to 
 * use a subclass instead. The most lightweight container class is "Sprite".</p>
 * 
 * <strong>Adding and removing children</strong>
 * 
 * <p>The class defines methods that allow you to add or remove children. When you add a child, 
 * it will be added at the frontmost position, possibly occluding a child that was added 
 * before. You can access the children via an index. The first child will have index 0, the 
 * second child index 1, etc.</p> 
 * 
 * Adding and removing objects from a container triggers non-bubbling events.
 * 
 * <ul>
 * <li><code>Event.ADDED</code>: the object was added to a parent.</li>
 * <li><code>Event.ADDED_TO_STAGE</code>: the object was added to a parent that is 
 * connected to the stage, thus becoming visible now.</li>
 * <li><code>Event.REMOVED</code>: the object was removed from a parent.</li>
 * <li><code>Event.REMOVED_FROM_STAGE</code>: the object was removed from a parent that 
 * is connected to the stage, thus becoming invisible now.</li>
 * </ul>
 * 
 * Especially the <code>ADDED_TO_STAGE</code> event is very helpful, as it allows you to 
 * automatically execute some logic (e.g. start an animation) when an object is rendered the 
 * first time.
 * 
 * @see Sprite
 * @see DisplayObject
 *  
 * @externs
 */
public class DisplayObjectContainer extends starling.display.DisplayObject {
	/**
	 *  @private 
	 */
	function DisplayObjectContainer() {
		super();
	}
	/**
	 *  Disposes the resources of all children. 
	 */
	override public function dispose():void {}
	/**
	 *  Adds a child to the container. It will be at the frontmost position. 
	 */
	public function addChild(child:starling.display.DisplayObject):starling.display.DisplayObject { return null; }
	/**
	 *  Adds a child to the container at a certain index. 
	 */
	public function addChildAt(child:starling.display.DisplayObject, index:int):starling.display.DisplayObject { return null; }
	/**
	 *  Removes a child from the container. If the object is not a child, the method returns
	 *      * <code>null</code>. If requested, the child will be disposed right away. 
	 */
	public function removeChild(child:starling.display.DisplayObject, dispose:Boolean = undefined):starling.display.DisplayObject { return null; }
	/**
	 *  Removes a child at a certain index. The index positions of any display objects above
	 *      * the child are decreased by 1. If requested, the child will be disposed right away. 
	 */
	public function removeChildAt(index:int, dispose:Boolean = undefined):starling.display.DisplayObject { return null; }
	/**
	 *  Removes a range of children from the container (endIndex included). 
	 *      * If no arguments are given, all children will be removed. 
	 */
	public function removeChildren(beginIndex:int = undefined, endIndex:int = undefined, dispose:Boolean = undefined):void {}
	/**
	 *  Returns a child object at a certain index. If you pass a negative index,
	 *      * '-1' will return the last child, '-2' the second to last child, etc. 
	 */
	public function getChildAt(index:int):starling.display.DisplayObject { return null; }
	/**
	 *  Returns a child object with a certain name (non-recursively). 
	 */
	public function getChildByName(name:String):starling.display.DisplayObject { return null; }
	/**
	 *  Returns the index of a child within the container, or "-1" if it is not found. 
	 */
	public function getChildIndex(child:starling.display.DisplayObject):int { return 0; }
	/**
	 *  Moves a child to a certain index. Children at and after the replaced position move up.
	 */
	public function setChildIndex(child:starling.display.DisplayObject, index:int):void {}
	/**
	 *  Swaps the indexes of two children. 
	 */
	public function swapChildren(child1:starling.display.DisplayObject, child2:starling.display.DisplayObject):void {}
	/**
	 *  Swaps the indexes of two children. 
	 */
	public function swapChildrenAt(index1:int, index2:int):void {}
	/**
	 *  Sorts the children according to a given function (that works just like the sort function
	 *      * of the Vector class). 
	 */
	public function sortChildren(compareFunction:Function):void {}
	/**
	 *  Determines if a certain object is a child of the container (recursively). 
	 */
	public function contains(child:starling.display.DisplayObject):Boolean { return false; }
	/**
	 *  @inheritDoc 
	 */
	override public function getBounds(targetSpace:starling.display.DisplayObject, out:openfl.geom.Rectangle = undefined):openfl.geom.Rectangle { return null; }
	/**
	 *  @inheritDoc 
	 */
	override public function hitTest(localPoint:openfl.geom.Point):starling.display.DisplayObject { return null; }
	/**
	 *  @inheritDoc 
	 */
	override public function render(painter:starling.rendering.Painter):void {}
	/**
	 *  Dispatches an event on all children (recursively). The event must not bubble. 
	 */
	public function broadcastEvent(event:starling.events.Event):void {}
	/**
	 *  Dispatches an event with the given parameters on all children (recursively). 
	 *      * The method uses an internal pool of event objects to avoid allocations. 
	 */
	public function broadcastEventWith(eventType:String, data:* = undefined):void {}
	/**
	 *  The number of children of this container. 
	 */
	public function get numChildren():int { return 0; }
	/**
	 *  If a container is a 'touchGroup', it will act as a single touchable object.
	 *      * Touch events will have the container as target, not the touched child.
	 *      * (Similar to 'mouseChildren' in the classic display list, but with inverted logic.)
	 *      * @default false 
	 */
	public function get touchGroup():Boolean { return false; }
	public function set touchGroup(value:Boolean):void {}
}
}
