package starling.events {
import starling.display.Stage;
import starling.display.DisplayObject;
/**
 *  The TouchProcessor is used to convert mouse and touch events of the conventional
 *  *  Flash stage to Starling's TouchEvents.
 *  *  
 *  *  <p>The Starling instance listens to mouse and touch events on the native stage. The
 *  *  attributes of those events are enqueued (right as they are happening) in the
 *  *  TouchProcessor.</p>
 *  *  
 *  *  <p>Once per frame, the "advanceTime" method is called. It analyzes the touch queue and
 *  *  figures out which touches are active at that moment; the properties of all touch objects
 *  *  are updated accordingly.</p>
 *  *  
 *  *  <p>Once the list of touches has been finalized, the "processTouches" method is called
 *  *  (that might happen several times in one "advanceTime" execution; no information is
 *  *  discarded). It's responsible for dispatching the actual touch events to the Starling
 *  *  display tree.</p>
 *  *  
 *  *  <strong>Subclassing TouchProcessor</strong>
 *  *  
 *  *  <p>You can extend the TouchProcessor if you need to have more control over touch and
 *  *  mouse input. For example, you could filter the touches by overriding the "processTouches"
 *  *  method, throwing away any touches you're not interested in and passing the rest to the
 *  *  super implementation.</p>
 *  *  
 *  *  <p>To use your custom TouchProcessor, assign it to the "Starling.touchProcessor"
 *  *  property.</p>
 *  *  
 *  *  <p>Note that you should not dispatch TouchEvents yourself, since they are
 *  *  much more complex to handle than conventional events (e.g. it must be made sure that an
 *  *  object receives a TouchEvent only once, even if it's manipulated with several fingers).
 *  *  Always use the base implementation of "processTouches" to let them be dispatched. That
 *  *  said: you can always dispatch your own custom events, of course.</p>
 *  
 * @externs
 */
public class TouchProcessor {
	/**
	 *  Creates a new TouchProcessor that will dispatch events to the given stage. 
	 */
	public function TouchProcessor(stage:starling.display.Stage) {}
	/**
	 *  Removes all event handlers on the stage and releases any acquired resources. 
	 */
	public function dispose():void {}
	/**
	 *  Analyzes the current touch queue and processes the list of current touches, emptying
	 *      * the queue while doing so. This method is called by Starling once per frame. 
	 */
	public function advanceTime(passedTime:Number):void {}
	/**
	 *  Enqueues a new touch or mouse event with the given properties. 
	 */
	public function enqueue(touchID:int, phase:String, globalX:Number, globalY:Number, pressure:Number = undefined, width:Number = undefined, height:Number = undefined):void {}
	/**
	 *  Enqueues an artificial touch that represents the mouse leaving the stage.
	 *      * 
	 *      * <p>On OS X, we get mouse events from outside the stage; on Windows, we do not.
	 *      * This method enqueues an artificial hover point that is just outside the stage.
	 *      * That way, objects listening for HOVERs over them will get notified everywhere.</p>
	 *      
	 */
	public function enqueueMouseLeftStage():void {}
	/**
	 *  Force-end all current touches. Changes the phase of all touches to 'ENDED' and
	 *      * immediately dispatches a new TouchEvent (if touches are present). Called automatically
	 *      * when the app receives a 'DEACTIVATE' event. 
	 */
	public function cancelTouches():void {}
	/**
	 *  Configures the margins within which, when a touch is starting, it's considered to be
	 *      *  a system gesture (in points). Note that you also need to enable 'ignoreSystemGestures'.
	 *      
	 */
	public function setSystemGestureMargins(topMargin:Number = undefined, bottomMargin:Number = undefined, leftMargin:Number = undefined, rightMargin:Number = undefined):void {}
	/**
	 *  Indicates if it multitouch simulation should be activated. When the user presses
	 *      * ctrl/cmd (and optionally shift), he'll see a second touch curser that mimics the first.
	 *      * That's an easy way to develop and test multitouch when there's only a mouse available.
	 *      
	 */
	public function get simulateMultitouch():Boolean { return false; }
	public function set simulateMultitouch(value:Boolean):void {}
	/**
	 *  The time period (in seconds) in which two touches must occur to be recognized as
	 *      * a multitap gesture. 
	 */
	public function get multitapTime():Number { return 0; }
	public function set multitapTime(value:Number):void {}
	/**
	 *  The distance (in points) describing how close two touches must be to each other to
	 *      * be recognized as a multitap gesture. 
	 */
	public function get multitapDistance():Number { return 0; }
	public function set multitapDistance(value:Number):void {}
	/**
	 *  The base object that will be used for hit testing. Per default, this reference points
	 *      * to the stage; however, you can limit touch processing to certain parts of your game
	 *      * by assigning a different object. 
	 */
	public function get root():starling.display.DisplayObject { return null; }
	public function set root(value:starling.display.DisplayObject):void {}
	/**
	 *  The stage object to which the touch objects are (per default) dispatched. 
	 */
	public function get stage():starling.display.Stage { return null; }
	/**
	 *  Returns the number of fingers / touch points that are currently on the stage. 
	 */
	public function get numCurrentTouches():int { return 0; }
	/**
	 *  If this callback returns <code>false</code>, the corresponding touch will have its
	 *      *  target set to <code>null</code>, which will prevent the original target from being
	 *      *  notified of the touch. In other words: the touch is being blocked. Callback format:
	 *      *  <pre>function(stageX:Number, stageY:Number):Boolean;</pre>
	 *      *  @default null
	 *      
	 */
	public function get occlusionTest():Function { return null; }
	public function set occlusionTest(value:Function):void {}
	/**
	 *  When enabled, all touches that start very close to the window edges are discarded.
	 *      *  On mobile, such touches often indicate swipes that are meant to open OS menus.
	 *      *  Per default, margins of 10 points at the very top and bottom of the screen are checked.
	 *      *  Call 'setSystemGestureMargins()' to adapt the margins in each direction.
	 *      *  @default true on mobile, false on desktop 
	 */
	public function get discardSystemGestures():Boolean { return false; }
	public function set discardSystemGestures(value:Boolean):void {}
}
}
