package starling.utils {
import starling.display.DisplayObject;
import openfl.geom.Point;
/**
 *  A utility class that can help with creating button-like display objects.
 *  *
 *  *  <p>When reacting to touch input, taps can easily be recognized through standard touch
 *  *  events via <code>TouchPhase.ENDED</code>. However, you often want a more elaborate kind of
 *  *  input handling, like that provide by Starling's <em>Button</em> class and its
 *  *  <em>TRIGGERED</em> event. It allows users to cancel a tap by moving the finger away from
 *  *  the object, for example; and it supports changing its appearance depending on its state.</p>
 *  *
 *  *  <p>Here is an example: a class that extends <em>TextField</em> and uses
 *  *  <em>ButtonBehavior</em> to add TRIGGER events and state-based coloring.</p>
 *  *
 *  *  <listing>
 *  *  public class TextButton extends TextField
 *  *  {
 *  *      private var _behavior:ButtonBehavior;
 *  *      private var _tint:uint = 0xffaaff;
 *  *      
 *  *      public function TextButton(width:int, height:int, text:String="",
 *  *                                 format:TextFormat=null, options:TextOptions=null)
 *  *      {
 *  *          super(width, height, text, format, options);
 *  *          _behavior = new ButtonBehavior(this, onStateChange);
 *  *      }
 *  *      
 *  *      private function onStateChange(state:String):void
 *  *      {
 *  *          if (state == ButtonState.DOWN) format.color = _tint;
 *  *          else format.color = 0xffffff;
 *  *      }
 *  *      
 *  *      public override function hitTest(localPoint:Point):DisplayObject
 *  *      {
 *  *          return _behavior.hitTest(localPoint);
 *  *      }
 *  *  }</listing>
 *  *
 *  *  <p>Instances of this class will now dispatch <em>Event.TRIGGERED</em> events (just like
 *  *  conventional buttons) and they will change their color when being touched.</p>
 *  
 * @externs
 */
public class ButtonBehavior {
	/**
	 *  Create a new ButtonBehavior.
	 * 	 *
	 * 	 * @param target           The object on which to listen for touch events.
	 * 	 * @param onStateChange    This callback will be executed whenever the button's state ought
	 * 	 *                         to change. <code>function(state:String):void</code>
	 * 	 * @param minHitAreaSize   If the display area of 'target' is smaller than a square of this
	 * 	 *                         size, its hit area will be extended accordingly.
	 * 	 * @param abortDistance    The distance you can move away your finger before triggering
	 * 	 *                         is aborted.
	 * 	 
	 */
	public function ButtonBehavior(target:starling.display.DisplayObject, onStateChange:Function, minHitAreaSize:Number = undefined, abortDistance:Number = undefined) {}
	/**
	 *  Forward your target's <code>hitTests</code> to this method to make sure that the hit
	 * 	 *  area is extended to <code>minHitAreaSize</code>. 
	 */
	public function hitTest(localPoint:openfl.geom.Point):starling.display.DisplayObject { return null; }
	/**
	 *  The current state of the button. The corresponding strings are found
	 * 	 *  in the ButtonState class. 
	 */
	public function get state():String { return null; }
	public function set state(value:String):void {}
	/**
	 *  The target on which this behavior operates. 
	 */
	public function get target():starling.display.DisplayObject { return null; }
	public function get_target():starling.display.DisplayObject { return null; }
	/**
	 *  The callback that is executed whenever the state changes.
	 * 	 *  Format: <code>function(state:String):void</code>
	 * 	 
	 */
	public function get onStateChange():Function { return null; }
	public function set onStateChange(value:Function):void {}
	public function get_onStateChange():Function { return null; }
	public function set_onStateChange(value:Function):Function { return null; }
	/**
	 *  Indicates if the mouse cursor should transform into a hand while it's over the button.
	 * 	 *  @default true 
	 */
	public function get useHandCursor():Boolean { return false; }
	public function set useHandCursor(value:Boolean):void {}
	public function get_useHandCursor():Boolean { return false; }
	public function set_useHandCursor(value:Boolean):Boolean { return false; }
	/**
	 *  Indicates if the button can be triggered. 
	 */
	public function get enabled():Boolean { return false; }
	public function set enabled(value:Boolean):void {}
	public function get_enabled():Boolean { return false; }
	public function set_enabled(value:Boolean):Boolean { return false; }
	/**
	 *  The target's hit area will be extended to have at least this width / height. 
	 * 	 *  Note that for this to work, you need to forward your hit tests to this class. 
	 */
	public function get minHitAreaSize():Number { return 0; }
	public function set minHitAreaSize(value:Number):void {}
	public function get_minHitAreaSize():Number { return 0; }
	public function set_minHitAreaSize(value:Number):Number { return 0; }
	/**
	 *  The distance you can move away your finger before triggering is aborted. 
	 */
	public function get abortDistance():Number { return 0; }
	public function set abortDistance(value:Number):void {}
	public function get_abortDistance():Number { return 0; }
	public function set_abortDistance(value:Number):Number { return 0; }
}
}
